/*
============================================================================
 Name        : GameControl.cpp
 Author      : Konstantin Tretyakov
 Version     :
 Copyright   : Copyright 2006, Konstantin Tretyakov
 Description : CGameControl implementation
============================================================================
*/

#include "GameControl.h"
#include "GameRenderer.h"
#include <e32math.h>

CGameControl::CGameControl(){
}


CGameControl::~CGameControl()
{
    delete iGameState;
    delete iSoundAdapter;
}

CGameControl* CGameControl::NewLC(EGLHelper* aEglHelper)
{
	CGameControl* self = new (ELeave) CGameControl;
    self->iEglHelper = aEglHelper;
	CleanupStack::PushL(self);
	self->ConstructL();
	return self;
}

CGameControl* CGameControl::NewL(EGLHelper* aEglHelper)
{
	CGameControl* self=CGameControl::NewLC(aEglHelper);
	CleanupStack::Pop(); // self;
	return self;
}

void CGameControl::ConstructL()
{
    iGameState = CGameState::NewL(this);
    iSoundAdapter = CSyMazeSoundAdapter::NewL();
}

void CGameControl::InitL(CEGLTextureHelper* textureHelper) {
    frame = 0;
    delete iGameState;
    iGameState = CGameState::NewL(this);
    iGameStatus = ETitle1Msg;
    iTextureHelper = textureHelper;
    InitRenderer(textureHelper);
}


void CGameControl::Cycle() {
    CGameState& gs = *iGameState;
    
    if (iGameStatus == EFirstPersonView || iGameStatus == EMapView) {
        frame++;
        gs.Behave();
    }

    RenderFrame(frame, gs, iGameStatus);
    iEglHelper->swapBuffers();    

    // To keep the background light on 
    if (!(frame%100)) User::ResetInactivityTime();

    // Suspend current thread to let other apps do smth
    if (!(frame%50)) User::After(0);

}

void CGameControl::Quit() {
    DestroyRenderer();
}


void CGameControl::Turn(int positive) {
    if (iGameStatus == EFirstPersonView || iGameStatus == EMapView)    
        iGameState->playerState.turningYaw += 4*positive;
}

void CGameControl::Move(int forward, int /*left*/, int /*up*/) {
    if (iGameStatus == EFirstPersonView || iGameStatus == EMapView)
        iGameState->playerState.movingForward += PLAYER_STEP_SIZE*forward;
        // Only support forward/backward
}

void CGameControl::MapToggle() {
    if (iGameStatus == EFirstPersonView) iGameStatus = EMapView;
    else if (iGameStatus == EMapView) iGameStatus = EFirstPersonView;
}

void CGameControl::Shoot() {
    if (iGameState->playerState.doingShot || !iGameState->playerState.canShoot) return;
    TReal cosYaw, sinYaw;
    Math::Sin(sinYaw, iGameState->playerState.yaw*3.1415/180);
    Math::Cos(cosYaw, iGameState->playerState.yaw*3.1415/180);
    iGameState->bullets.bullets.Append(
        TPointObject(iGameState->playerState.posX, 
                     iGameState->playerState.posY, 
                     iGameState->playerState.posZ + 0.3,
                     0.4*sinYaw, 0.4*cosYaw, 0));

    iGameState->playerState.shotFrame = 0;
    iGameState->playerState.doingShot = TRUE;
    iSoundAdapter->PlayL(); // OK, I'm spoiling the convention here.
                            // Shoot should actually be called ShootL because of this
                            // sound here. But sound is secondary anyway, so we don't care.
}

void CGameControl::OKButton() {
    switch (iGameStatus) {
        case ETitle1Msg:
            iGameStatus = ETitle2Msg;
            break;
        case ETitle2Msg:
            iGameStatus = ETitle3Msg;
            break;
        case ETitle3Msg:
        case EBonusMapMsg:
        case EBonusPortalMsg:
        case EBonusWeaponMsg:        
            iGameStatus = EFirstPersonView;
            break;
        case EFinishMsg:
            // Restart game
            InitL(iTextureHelper);
            break;
        case EFirstPersonView:
        case EMapView:
            // Switch FirstPerson/Map, if allowed
            if (iGameState->playerState.canViewMap) MapToggle();
            break;
    }
}


void CGameControl::ReachedLevel1(){
    iGameStatus = EBonusMapMsg;
}

void CGameControl::ReachedLevel2(){
    iGameStatus = EBonusPortalMsg;
}

void CGameControl::ReachedFinish(){
    iGameStatus = EFinishMsg;
}

void CGameControl::FoundWeapon(){
    iGameStatus = EBonusWeaponMsg;
}

