/*
============================================================================
 Name        : GameWorld.cpp
 Author      : Konstantin Tretyakov
 Version     :
 Copyright   : Copyright 2006, Konstantin Tretyakov
 Description : Various objects in the game
============================================================================
*/

#include "GameWorld.h"
#include "GameControl.h"
#include <e32math.h>

// --------------------------------------------------------------- //
/**
 * Convenience routine
 */
void normalize(float& x, float& y, float& z) {
    double d2 = x*x + y*y + z*z;
    double d;
    Math::Sqrt(d, d2);
    x /= d;
    y /= d;
    z /= d;
}


// --------------------------------------------------------------- //
/**
 * Player behaviour: moving + collision detection
 */
void TPlayerState::Behave(TLabyrinth& maze) {
    // Move player    
    TReal cosYaw, sinYaw;
    Math::Cos(cosYaw, yaw*3.1415/180);
    Math::Sin(sinYaw, yaw*3.1415/180);

    TReal newPosX, newPosY;
    
    // Collision detection
    TReal signForward = movingForward;
    bool collided = TRUE;
    while (collided) {
        newPosX = posX + sinYaw*signForward;
        newPosY = posY + cosYaw*signForward;
        if (maze((int)(newPosX + PLAYER_RADIUS),(int)newPosY) > 0 ||
            maze((int)(newPosX - PLAYER_RADIUS),(int)newPosY) > 0 ||
            maze((int)(newPosX),(int)(newPosY + PLAYER_RADIUS)) > 0 ||
            maze((int)(newPosX),(int)(newPosY - PLAYER_RADIUS)) > 0) {
            signForward /= 2;
            if (signForward < PLAYER_RADIUS && signForward > -PLAYER_RADIUS)
                signForward = 0;
        }
        else collided = FALSE;
    }

    posX = newPosX;
    posY = newPosY;    
    yaw  += turningYaw;
    
    // Friction
    movingForward *= 0.8;
    turningYaw    *= 0.8;
    
    if (movingForward < D_TOLERANCE && movingForward > -D_TOLERANCE) movingForward = 0;
    if (turningYaw  < D_TOLERANCE && turningYaw > -D_TOLERANCE) turningYaw = 0;
    
    // Shot animation?
    if (doingShot) {
        shotFrame++;
        if (shotFrame == SHOT_FRAME_COUNT) doingShot = FALSE;
    }
    
    // "Wow" animation?
    if (wowed) wowed = FALSE;
}


// --------------------------------------------------------------- //

CGameState::CGameState():
    boids(200, 0, MAZE_WIDTH+1, 0, MAZE_HEIGHT+1, 1, 4),
    maze()
{
    // Initialize random maze
    maze.MakeRandom();
    
    // Add some random "squares"
    maze.AddRandomSquare(5, 5);
    maze.AddRandomSquare(4, 5);
    maze.AddRandomSquare(3, 3);
    maze.AddRandomSquare(2, 4);
    maze.AddRandomSquare(3, 1);
    
    // Make some way forward (so that user won't be looking into the wall when starting)
    maze(1, 2) = 0; 
    maze(1, 3) = 0;
    maze(1, 4) = 0;
    maze(1, 5) = 0;


    // Mark start position
    maze(1, 1) = -2;
    
    int wx = (Math::Random() % (MAZE_WIDTH - 2)) + 1;
    int wy = (Math::Random() % (MAZE_HEIGHT - 2)) + 1;
    maze(wx, wy) = 0;
    thingy = TMovingThingy((float)wx + 0.5, (float)wy + 0.5, 0.5);    
}

// --------------------------------------------------------------- //
CGameState::~CGameState()
{
}

// --------------------------------------------------------------- //
CGameState* CGameState::NewLC(CGameControl* control)
{
    CGameState* self = new (ELeave) CGameState;
    CleanupStack::PushL(self);
    self->ConstructL(control);
    return self;
}

// --------------------------------------------------------------- //
CGameState* CGameState::NewL(CGameControl* control)
{
    CGameState* self=CGameState::NewLC(control);
    CleanupStack::Pop(); // self;
    return self;
}

// --------------------------------------------------------------- //
void CGameState::ConstructL(CGameControl* control)
{
    this->iGameControl = control;
}


// --------------------------------------------------------------- //
/**
 * Game behaviour:
 *      player, thingy, bullets, boids
 */
void CGameState::Behave() {
    playerState.Behave(maze);
    thingy.Behave();
    bullets.Behave(maze);
    boids.Behave();
    drops.Behave();
    
    // Boids generate drops
    if (Math::Random() % 2 == 0) {
        TPointObject& b = boids.boids[Math::Random() % boids.boids.Count()];
        drops.drops.Append(TPointObject(b.posX, b.posY, b.posZ, 0, 0, 0));
    }
    
    // If any of the drops touches a player, react
    for (int i = 0; i < drops.drops.Count(); i++) {
        TPointObject& d = drops.drops[i];
        if (d.alive) {
            #define abs(x) ((x >= 0) ? (x) : -(x));
            float dx = abs(d.posX - playerState.posX);
            float dy = abs(d.posY - playerState.posY);
            float dz = abs(d.posZ - 0.5);
            if (dx < PLAYER_INTERACTION_RADIUS && dy < PLAYER_INTERACTION_RADIUS && dz < 0.5) {

                // Yes, we've caught a drop
                playerState.wowed = TRUE;
                d.alive = FALSE;
                playerState.dropsCaught++;
                
                if (playerState.dropsCaught == DROPS_TO_CATCH_1) {
                    // Reached level 1. Allow to use Map.
                    playerState.canViewMap = TRUE;
                    iGameControl->ReachedLevel1();
                }
                else if (playerState.dropsCaught == DROPS_TO_CATCH_2){
                    // Reached level 2. Open exit
                    maze(MAZE_WIDTH - 2, MAZE_HEIGHT -2) = -3;
                    maze(MAZE_WIDTH - 1, MAZE_HEIGHT - 2) = -3; // Make exit
                    iGameControl->ReachedLevel2();                 
                }
                break;
            }
        }
    }
    
    // Did the player find the thingy?
    if (thingy.alive) {
        float dx = abs(thingy.posX - playerState.posX);
        float dy = abs(thingy.posY - playerState.posY);
        float dz = abs(thingy.posZ - 0.5);
        if (dx < PLAYER_INTERACTION_RADIUS && dy < PLAYER_INTERACTION_RADIUS && dz < 0.5) {
            playerState.canShoot = TRUE;
            thingy.alive = FALSE;            
            iGameControl->FoundWeapon();
        }
    }
    
    
    // Did the player reach exit?
    if (playerState.posX > MAZE_WIDTH || playerState.posY > MAZE_HEIGHT) {
        iGameControl->ReachedFinish();
    }
}


// --------------------------------------------------------------- //
/**
 * Moving thingy does nothing
 */
void TMovingThingy::Behave() {
    frame++;
}


// --------------------------------------------------------------- //
/**
 * Bullets move. When a bullet hits the wall it dies.
 */
void TBullets::Behave(TLabyrinth& maze) {

    // Moving & collision detection for bullets
    for (int i = 0; i < bullets.Count(); i++) {
        TPointObject& b = bullets[i];
        b.frame++;
        if (b.alive) {
            int& m = maze((int)b.posX, (int)b.posY);
            if (m > 0 ||
                b.posX <= -MAZE_WIDTH || b.posX >= 2*MAZE_WIDTH || 
                b.posY <= -MAZE_WIDTH || b.posY >= 2*MAZE_HEIGHT ||
                (b.dX == 0 && b.dY == 0 && b.dZ == 0)) {
                // Bump!
                b.alive = FALSE;
                
                // Damage the cube
                // XXX: The following are related to the mazeColors in GameRendererStaticData.h, careful!
                // TODO: Bad code
                #define CUBE_DAMAGE_0 4
                #define CUBE_DAMAGE_MAX 9
                if (m <= CUBE_DAMAGE_0) m = CUBE_DAMAGE_0;
                m++;
                if (m == CUBE_DAMAGE_MAX) {
                    m = -CUBE_DAMAGE_0 - 3;
                    if ((int)b.posX == 0 || (int)b.posX == MAZE_WIDTH - 1 ||
                        (int)b.posY == 0 || (int)b.posY == MAZE_HEIGHT - 1)
                        m = CUBE_DAMAGE_MAX - 1;
                }
                    
            }
            b.posX += b.dX;
            b.posY += b.dY;
            b.posZ += b.dZ;
        }
    }

    // Clean dead bullets
    for (int i = 0; i < bullets.Count(); i++) {
        if (!bullets[i].alive) bullets.Remove(i);
    }    
}

TBullets::~TBullets(){
    bullets.Reset();
}


// --------------------------------------------------------------- //
/**
 * Drops drop. When a drop hits bottom it dies
 */
void TDrops::Behave() {

    // Moving & collision detection for drops
    for (int i = 0; i < drops.Count(); i++) {
        TPointObject& b = drops[i];
        b.frame++;      
        b.posX += b.dX;
        b.posY += b.dY;
        b.posZ += b.dZ;
        
        // Acceleration
        b.dZ += -0.01;
        
        if (b.posZ < 0) b.alive = FALSE;
    }

    // Clean dead drops
    for (int i = 0; i < drops.Count(); i++) {
        if (!drops[i].alive) drops.Remove(i);
    }    
}

TDrops::~TDrops(){
    drops.Reset();
}


// --------------------------------------------------------------- //
/**
 * Initializes a random set of boids distributed uniformly within [0, maxX] x [0, maxY] x [0, maxZ]
 */
TBoids::TBoids(int num, float minX, float maxX, float minY, float maxY, float minZ, float maxZ):
    minX(minX), minY(minY), minZ(minZ),maxX(maxX), maxY(maxY), maxZ(maxZ) {
    // Generate num random boids
    for (int i = 0; i < num; i++) {
        float rndX = (float)(Math::Random()%1000)*(maxX-minX)/1000 + minX;
        float rndY = (float)(Math::Random()%1000)*(maxY-minY)/1000 + minY;
        float rndZ = (float)(Math::Random()%1000)*(maxZ-minZ)/1000 + minZ;
        float rndDX = (float)(Math::Random()%1000)/10000 - 0.05; // -0.05..0.05
        float rndDY = (float)(Math::Random()%1000)/10000 - 0.05;
        float rndDZ = (float)(Math::Random()%1000)/10000 - 0.05;
        boids.Append(TPointObject(rndX, rndY, rndZ, rndDX, rndDY, rndDZ));
    }
}

TBoids::~TBoids() {
    boids.Reset();
}

/**
 * Boids flock (separation, alignment, cohesion).
 * If they move over the [0, max] rectangle they wrap.
 */
void TBoids::Behave() {
    // Move boids
    for (int i = 0; i < boids.Count(); i++) {
        TPointObject& b = boids[i];
        b.posX += b.dX;
        b.posY += b.dY;
        b.posZ += b.dZ;
        if (b.posX > maxX) b.posX = minX;
        if (b.posX < minX) b.posX = maxX;
        if (b.posY > maxY) b.posY = minY;
        if (b.posY < minY) b.posY = maxY;
        if (b.posZ > maxZ) b.posZ = minZ;
        if (b.posZ < minZ) b.posZ = maxZ;        
    }
    
    // Flocking behaviour
    RArray<int> neighbors;
    RArray<float> neighborsd2;
    
    #define DETECTION_RANGE 1
    #define SEPARATION_RANGE 0.3
    
    for (int i = 0; i < boids.Count(); i++) {
        TPointObject& b = boids[i];
        
        // Select neighboring boids
        for (int j = 0; j < boids.Count(); j++) {
            if (j != i) {
                float distx = b.posX - boids[j].posX;
                float disty = b.posY - boids[j].posY;
                float distz = b.posZ - boids[j].posZ;
                float d2 = distx*distx + disty*disty + distz*distz;
                if (d2 < DETECTION_RANGE*DETECTION_RANGE) {
                    neighbors.Append(j);
                    neighborsd2.Append(d2);
                }
            }
        }
        
        // Separation: steer to avoid crowding if too close
        float steerdx = 0, steerdy = 0, steerdz = 0;
        int num = 0;
        for (int j = 0; j < neighbors.Count(); j++) {
            if (neighborsd2[j] < SEPARATION_RANGE*SEPARATION_RANGE) {
                num++;
                steerdx += b.posX - boids[neighbors[j]].posX;
                steerdy += b.posY - boids[neighbors[j]].posY;
                steerdz += b.posZ - boids[neighbors[j]].posZ;
            }
        }
        if (num > 0) normalize(steerdx, steerdy, steerdz);        
        
        // Alignment: align speed to group average
        float aligndx = 0, aligndy = 0, aligndz = 0;
        num = 0;
        for (int j = 0; j < neighbors.Count(); j++) {
            num++;
            aligndx += boids[neighbors[j]].dX;
            aligndy += boids[neighbors[j]].dY;
            aligndz += boids[neighbors[j]].dZ;
        }
        
        if (num > 0) {
            aligndx /= num;
            aligndy /= num;
            aligndz /= num;
            
            aligndx -= b.dX;
            aligndy -= b.dY;
            aligndz -= b.dZ;
            normalize(aligndx, aligndy, aligndz);
        }
        
        // Cohesion: steer closer if far away
        float cohesdx = 0, cohesdy = 0, cohesdz = 0;
        num = 0;
        for (int j = 0; j < neighbors.Count(); j++) {
            if (neighborsd2[j] >= SEPARATION_RANGE*SEPARATION_RANGE) {
                num++;
                cohesdx += boids[neighbors[j]].posX - b.posX;
                cohesdy += boids[neighbors[j]].posY - b.posY;
                cohesdz += boids[neighbors[j]].posZ - b.posZ;
            }
        }
        if (num > 0) normalize(cohesdx, cohesdy, cohesdz);
        
        // Combine the modification and apply to boid (btw this is not very correct,
        // as it is immediately used for other boids)
        b.dX += (0.1*cohesdx + 1*steerdx + 0.3*aligndx)/10;
        b.dY += (0.1*cohesdy + 1*steerdy + 0.3*aligndy)/10;
        b.dZ += (0.1*cohesdz + 1*steerdz + 0.3*aligndz)/10;
        
        normalize(b.dX, b.dY, b.dZ);
        b.dX /= 10;
        b.dY /= 10;
        b.dZ /= 10;
        
        // Cleanup for next boid
        neighbors.Reset(); neighborsd2.Reset();
    }
    
    neighbors.Reset(); // Just in case..
    neighborsd2.Reset();
}



